@@ -111,7 +111,8 @@ static gint               xfce_appfinder_model_item_compare_frecency  (gconstpoi
 static void               xfce_appfinder_model_frecency_collect       (XfceAppfinderModel       *model,
                                                                        GMappedFile              *mmap);
 static void               xfce_appfinder_model_frecency_free          (gpointer                  data);
-static void               xfce_appfinder_model_migrate_history_file   (void);
+
+
 
 struct _XfceAppfinderModelClass
 {
@@ -975,6 +976,21 @@ xfce_appfinder_model_item_compare_command (gconstpointer a,
 
 
 
+static gint
+xfce_appfinder_model_item_compare_command_strict (gconstpointer a,
+                                                  gconstpointer b)
+{
+  const ModelItem *item_a = a, *item_b = b;
+
+  /* ignore non custom commands (i.e. garcon menu items) */
+  if (item_a->item != NULL || item_b->item != NULL)
+    return -1;
+
+  return g_utf8_collate (item_a->command, item_b->command);
+}
+
+
+
 static gchar *
 xfce_appfinder_model_item_key (GarconMenuItem *item)
 {
@@ -1221,7 +1237,8 @@ xfce_appfinder_model_file_get_mtime (GFile *file)
 
 static gboolean
 xfce_appfinder_model_history_insert (XfceAppfinderModel *model,
-                                     const gchar        *command)
+                                     const gchar        *command,
+                                     ModelItem         **model_item)
 {
   ModelItem    *item;
   GtkTreeIter   iter;
@@ -1235,12 +1252,23 @@ xfce_appfinder_model_history_insert (XfceAppfinderModel *model,
   /* add new command */
   item = g_slice_new0 (ModelItem);
   item->command = g_strdup (command);
-  if (g_slist_find_custom (model->items, item, xfce_appfinder_model_item_compare_command) != NULL)
+
+  /* check if command matches any existing item, including the ones still being loaded (collect_items) */
+  if (g_slist_find_custom (model->items, item, xfce_appfinder_model_item_compare_command) != NULL ||
+      g_slist_find_custom (model->collect_items, item, xfce_appfinder_model_item_compare_command) != NULL)
     {
-       APPFINDER_DEBUG ("Skip adding %s to the model as it's already contained.", command);
-       g_slice_free (ModelItem, item);
-       return FALSE;
+      APPFINDER_DEBUG ("Command %s already exists, mark as not visible.", command);
+      item->not_visible = TRUE;
     }
+
+  /* check if command matches any existing item which is strictly command item (i.e. not from garcon) */
+  if (g_slist_find_custom (model->items, item, xfce_appfinder_model_item_compare_command_strict) != NULL)
+    {
+      APPFINDER_DEBUG ("Skip adding %s to the history as it's already present.", command);
+      g_slice_free (ModelItem, item);
+      return FALSE;
+    }
+
   item->surface = cairo_surface_reference (model->command_surface);
   item->surface_large = cairo_surface_reference (model->command_surface_large);
   model->items = g_slist_insert_sorted (model->items, item, xfce_appfinder_model_item_compare);
@@ -1260,6 +1288,10 @@ xfce_appfinder_model_history_insert (XfceAppfinderModel *model,
   gtk_tree_path_free (path);
 
   g_hash_table_insert (model->items_hash, item->command, item);
+
+  if (model_item)
+    *model_item = item;
+
   return TRUE;
 }
 
@@ -1314,7 +1346,7 @@ xfce_appfinder_model_history_changed (GFileMonitor       *monitor,
                         {
                           /* look for new commands */
                           command = g_strndup (contents, end - contents);
-                          xfce_appfinder_model_history_insert (model, command);
+                          xfce_appfinder_model_history_insert (model, command, NULL);
                           g_free (command);
                         }
                       contents = end + 1;
@@ -1628,10 +1660,9 @@ xfce_appfinder_model_collect_history (XfceAppfinderModel *model,
 
       if (end != contents)
         {
-          item = g_slice_new0 (ModelItem);
-          item->command = g_strndup (contents, end - contents);
-          item->surface = cairo_surface_reference (model->command_surface);
-          item->surface_large = cairo_surface_reference (model->command_surface_large);
+          gchar *command = g_strndup (contents, end - contents);
+          xfce_appfinder_model_history_insert (model, command, &item);
+          g_free (command);
           model->collect_items = g_slist_prepend (model->collect_items, item);
         }
 
@@ -2017,8 +2048,6 @@ xfce_appfinder_model_collect_thread (gpointer user_data)
         }
     }
 
-  xfce_appfinder_model_migrate_history_file ();
-
   /* load command history */
   filename = xfce_resource_lookup (XFCE_RESOURCE_CACHE, NEW_HISTORY_PATH);
   if (G_LIKELY (filename != NULL))
@@ -2209,44 +2238,6 @@ xfce_appfinder_model_frecency_free (gpointer data)
 
 
 
-/* Migration for history cache file, can be removed during the 4.20 cycle */
-static void
-xfce_appfinder_model_migrate_history_file (void)
-{
-  gchar  *filename;
-  GFile  *old_file, *new_file, *dir;
-  GError *error = NULL;
-
-  filename = xfce_resource_lookup (XFCE_RESOURCE_CACHE, OLD_HISTORY_PATH);
-
-  if (filename == NULL)
-    return;
-
-  old_file = g_file_new_for_path (filename);
-  g_free (filename);
-
-  new_file = g_file_resolve_relative_path (old_file, "../../../" NEW_HISTORY_PATH);
-
-  dir = g_file_get_parent (new_file);
-  g_file_make_directory_with_parents (dir, NULL, NULL);
-  g_object_unref (dir);
-
-  if (g_file_query_exists (new_file, NULL))
-    g_warning ("Old and new history files exist, migration aborted");
-  else if (g_file_move (old_file, new_file, G_FILE_COPY_NONE, NULL, NULL, NULL, &error))
-    g_message ("Migration of history file was successful");
-  else
-    {
-      g_warning ("Failed to migrate history file: %s", error->message);
-      g_clear_error (&error);
-    }
-
-  g_object_unref (old_file);
-  g_object_unref (new_file);
-}
-
-
-
 guint
 xfce_appfinder_model_calculate_frecency (guint   frequency,
                                          guint64 recency)
@@ -2521,13 +2512,13 @@ xfce_appfinder_model_get_visible_command (XfceAppfinderModel *model,
 
   item = ITER_GET_DATA (iter);
 
-  if (item->item != NULL)
-    {
-      appfinder_return_val_if_fail (GARCON_IS_MENU_ITEM (item->item), FALSE);
-
-      if (item->not_visible)
-        return FALSE;
-    }
+  /*
+   * besides hidden garcon menus, a command may be not visible if it's
+   * a duplicate of an existing menu. For example, if "xfce4-about" is
+   * in the command history, it should not be displayed twice.
+   */
+  if (item->not_visible)
+    return FALSE;
 
   if (item->command != NULL && string != NULL)
     return xfce_appfinder_model_fuzzy_match (item->command, string);
@@ -2724,7 +2715,7 @@ xfce_appfinder_model_save_command (XfceAppfinderModel  *model,
     return TRUE;
 
   /* add command to the model history if unique, else just return */
-  if (!xfce_appfinder_model_history_insert (model, command))
+  if (!xfce_appfinder_model_history_insert (model, command, NULL))
     return TRUE;
 
   /* add to the hashtable */
