@@ -57,6 +57,7 @@
 
 
 static void       xfce_appfinder_window_finalize                      (GObject                     *object);
+static void       xfce_appfinder_window_focus_out                     (GtkWidget                   *widget);
 static void       xfce_appfinder_window_unmap                         (GtkWidget                   *widget);
 static gboolean   xfce_appfinder_window_key_press_event               (GtkWidget                   *widget,
                                                                        GdkEventKey                 *event);
@@ -93,12 +94,15 @@ static void       xfce_appfinder_window_drag_data_get                 (GtkWidget
 static gboolean   xfce_appfinder_window_treeview_key_press_event      (GtkWidget                   *widget,
                                                                        GdkEventKey                 *event,
                                                                        XfceAppfinderWindow         *window);
+static gboolean   xfce_appfinder_window_completion_match_selected     (GtkEntryCompletion          *widget,
+                                                                       GtkTreeModel                *model,
+                                                                       GtkTreeIter                 *iter,
+                                                                       gpointer                     data);
 static void       xfce_appfinder_window_category_changed              (GtkTreeSelection            *selection,
                                                                        XfceAppfinderWindow         *window);
 static void       xfce_appfinder_window_category_set_categories       (XfceAppfinderModel          *model,
                                                                        XfceAppfinderWindow         *window);
-static void       xfce_appfinder_window_preferences                   (GtkWidget                   *button,
-                                                                       XfceAppfinderWindow         *window);
+static void       xfce_appfinder_window_preferences                   (XfceAppfinderWindow         *window);
 static void       xfce_appfinder_window_property_changed              (XfconfChannel               *channel,
                                                                        const gchar                 *prop,
                                                                        const GValue                *value,
@@ -165,10 +169,10 @@ struct _XfceAppfinderWindow
   GarconMenuDirectory        *filter_category;
   gchar                      *filter_text;
 
-  guint                       idle_entry_changed_id;
-
   gint                        last_window_height;
 
+  guint                       idle_entry_changed_id;
+  gulong                      focus_out_id;
   gulong                      property_watch_id;
   gulong                      categories_changed_id;
 };
@@ -303,6 +307,8 @@ xfce_appfinder_window_init (XfceAppfinderWindow *window)
   g_object_set (G_OBJECT (completion), "text-column", XFCE_APPFINDER_MODEL_COLUMN_COMMAND, NULL);
   gtk_entry_completion_set_popup_single_match (completion, TRUE);
   gtk_entry_completion_set_inline_completion (completion, TRUE);
+  g_signal_connect (G_OBJECT (completion), "match-selected",
+      G_CALLBACK (xfce_appfinder_window_completion_match_selected), window);
 
   cell = gtk_cell_renderer_text_new ();
   g_object_set (G_OBJECT (cell), "ellipsize", PANGO_ELLIPSIZE_END, NULL);
@@ -375,7 +381,7 @@ xfce_appfinder_window_init (XfceAppfinderWindow *window)
 
   button = gtk_button_new_with_mnemonic (_("_Preferences"));
   gtk_box_pack_start (GTK_BOX (hbox), button, FALSE, FALSE, 0);
-  g_signal_connect (G_OBJECT (button), "clicked",
+  g_signal_connect_swapped (G_OBJECT (button), "clicked",
       G_CALLBACK (xfce_appfinder_window_preferences), window);
   gtk_widget_set_valign (button, GTK_ALIGN_CENTER);
   gtk_button_set_always_show_image(GTK_BUTTON(button), TRUE);
@@ -417,6 +423,10 @@ xfce_appfinder_window_init (XfceAppfinderWindow *window)
     g_signal_connect (G_OBJECT (window->channel), "property-changed",
         G_CALLBACK (xfce_appfinder_window_property_changed), window);
 
+  /* check if window should be closed on focus lost */
+  window->focus_out_id =
+    g_signal_connect (window, "focus-out-event", G_CALLBACK (xfce_appfinder_window_focus_out), NULL);
+
   APPFINDER_DEBUG ("constructed window");
 }
 
@@ -462,6 +472,17 @@ xfce_appfinder_window_finalize (GObject *object)
 
 
 
+static void
+xfce_appfinder_window_focus_out (GtkWidget *widget)
+{
+  XfceAppfinderWindow *window = XFCE_APPFINDER_WINDOW (widget);
+
+  if (xfconf_channel_get_bool (window->channel, "/close-on-focus-lost", FALSE))
+    gtk_window_close (GTK_WINDOW (widget));
+}
+
+
+
 static void
 xfce_appfinder_window_unmap (GtkWidget *widget)
 {
@@ -1667,6 +1688,30 @@ xfce_appfinder_window_treeview_key_press_event (GtkWidget           *widget,
 
 
 
+static gboolean
+xfce_appfinder_window_completion_match_selected (GtkEntryCompletion *widget,
+                                                 GtkTreeModel       *model,
+                                                 GtkTreeIter        *iter,
+                                                 gpointer            data)
+{
+  GdkModifierType      state;
+  XfceAppfinderWindow *window = data;
+  gchar               *cmd = NULL;
+
+  gtk_get_current_event_state (&state);
+
+  if (state & GDK_CONTROL_MASK) {
+    gtk_tree_model_get (GTK_TREE_MODEL (window->model), iter, XFCE_APPFINDER_MODEL_COLUMN_COMMAND, &cmd, -1);
+    gtk_entry_set_text (GTK_ENTRY (window->entry), cmd);
+    xfce_appfinder_window_execute (window, TRUE);
+    return TRUE;
+  }
+
+  return FALSE;
+}
+
+
+
 static void
 xfce_appfinder_window_entry_icon_released (GtkEntry             *entry,
                                            GtkEntryIconPosition  icon_pos,
@@ -1769,16 +1814,13 @@ xfce_appfinder_window_category_set_categories (XfceAppfinderModel  *signal_from_
 
 
 static void
-xfce_appfinder_window_preferences (GtkWidget           *button,
-                                   XfceAppfinderWindow *window)
+xfce_appfinder_window_preferences (XfceAppfinderWindow *window)
 {
-  appfinder_return_if_fail (GTK_IS_WIDGET (button));
-
   /* preload the actions, to make sure there are default values */
   if (window->actions == NULL)
     window->actions = xfce_appfinder_actions_get ();
 
-  xfce_appfinder_preferences_show (gtk_widget_get_screen (button));
+  xfce_appfinder_preferences_show (window);
 }
 
 
@@ -2003,6 +2045,15 @@ xfce_appfinder_window_launch_clicked (XfceAppfinderWindow *window)
 
 
 
+static gboolean
+xfce_appfinder_window_destroy_widget (gpointer data)
+{
+  gtk_widget_destroy (GTK_WIDGET (data));
+  return FALSE;
+}
+
+
+
 static void
 xfce_appfinder_window_execute (XfceAppfinderWindow *window,
                                gboolean             close_on_succeed)
@@ -2086,7 +2137,7 @@ xfce_appfinder_window_execute (XfceAppfinderWindow *window,
     }
 
   if (result && close_on_succeed)
-    gtk_widget_destroy (GTK_WIDGET (window));
+    g_idle_add (xfce_appfinder_window_destroy_widget, window);
 }
 
 
@@ -2117,6 +2168,22 @@ xfce_appfinder_window_update_frecency (XfceAppfinderWindow *window,
 
 
 
+/*
+ * When 'close-on-focus-lost' is enabled, calling this function prevents
+ * the main window being closed when preferences is opened.
+ */
+void
+xfce_appfinder_window_keep_open (XfceAppfinderWindow *window,
+                                 gboolean             keep_open)
+{
+  if (keep_open)
+    g_signal_handler_block (window, window->focus_out_id);
+  else
+    g_signal_handler_unblock (window, window->focus_out_id);
+}
+
+
+
 void
 xfce_appfinder_window_set_expanded (XfceAppfinderWindow *window,
                                     gboolean             expanded)
